//================================================================================================
//=  HGT-DETECTION v3.2b
//=  Authors : Alix Boc and Vladimir Makarenkov
//=  Date : November 2009
//=
//=  Description : This program detect horizontal gene transfer (HGT). As input it takes 2 
//=  trees: a species tree and a gene tree. the goal is to transform the species tree
//=  into the gene tree following a transfer scenario. There are 3 criteria : the robinson and
//=  Foulds distance, the least-square criterion and the bipartition distance. We also use the
//=  subtree constraint. With this version we can now perform simulation.
//=
//=	 input   : file with species tree and gene tree in the newick format.
//=			   In case of simulation, the species tree and all the gene trees in the same file in
//=            the phylip format or newick string
//=  output  : a list of HGT and the criteria values for each one.
//=	 options :
//=
//================================================================================================


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <signal.h>

#pragma warning(disable:4996)

#include "structures.h"
#include "utils_tree.cpp"
#include "fonctions.cpp"

#define binaireSpecies 0 
#define binaireGene    1

void traiterSignal(int sig){
	printf("\nMESSAGE : SEGMENTATION FAULT #%d DETECTED",sig);
	printf("\nUse valgrind or gdb to fix the problem");
	printf("\n");
	exit(-1);
}

//========================================================================================================
//============================================ MAIN ======================================================
//========================================================================================================
int main(int nargc,char **argv){

	struct InputTree SpeciesTree;				    //== initial species tree
	struct InputTree SpeciesTreeCurrent;		//== initial species tree
	struct InputTree FirstTree;
	struct InputTree GeneTree;					    //== initial gene tree
	struct InputTree SpeciesTreeRed;			  //== reduced species tree
	struct InputTree GeneTreeRed;				    //== reduced gene tree
	struct ReduceTrace aMap;					      //== mapping structure between species tree and recuded species tree
	struct InputTree geneTreeSave;
	struct HGT * bestHGTRed = NULL;				  //== list of HGT for the reduced tree
	struct HGT * bestHGT = NULL;				    //== list of HGT for the normale tree
	struct HGT * outHGT = NULL;
	struct HGT * bestHGTmulticheck = NULL;
	int nbHGT_boot;
	int first = 1,k,l;
	int cpt_hgt,i,j,tmp,nbTree=0;
	int bootstrap = 0;
	int multigene = 0;
	int nbHgtFound = 0;
	struct CRITERIA * multicheckTab=NULL;
	struct CRITERIA aCrit;						       //== struture of all the criteria
	struct DescTree *DTSpecies,					       //== structure of submatrices for the species tree
					*DTGene;					       //== structure of submatrices for the gene tree
	struct Parameters param;
	FILE *in,*out;
	int max_hgt,nbHGT;
	int ktSpecies;
    int trivial = 1;
	int *speciesLeaves = NULL;
	int RFref;
	int imc;	
	char *mot = (char*)malloc(100);
	
	//srand(2345);
	initInputTree(&geneTreeSave);


	//== read parameters
	printf("\nhgt : reading options");
	if(readParameters(&param,argv,nargc)==-1){
		printf("\nhgt : no options specified, see the README file for more details\n");
		exit(-1);
	}

	rand_bootstrap = param.rand_bootstrap;
	
	//if(strcmp(param.version,"consol")==0)
		signal(SIGSEGV,traiterSignal);

	//if(strcmp(param.version,"consol")==0)
	//	printf(startMessage);

	//== open the output file
	//if((out=fopen(param.outputfile,"w+"))==NULL){
	//	printf("Can't open output file (%s)",param.outputfile);
	//	exit(-1);
	//}

	//PrintHeader(out,param);
	
	//== open the input file
	if((in=fopen(param.inputfile,"r"))==NULL){
		printf("\nhgt : The file %s does not exist",param.inputfile);
		exit(-1);
	}
	if(strcmp(param.speciesroot,"file") == 0){
		if(!file_exists(param.speciesRootfileLeaves)){
			printf("\nhgt : The file %s does not exist",param.speciesRootfileLeaves);
			exit(-1);
		}
	}
	if(strcmp(param.generoot,"file") == 0){
		if(!file_exists(param.geneRootfileLeaves)){
			printf("\nhgt : The file %s does not exist",param.geneRootfileLeaves);
			exit(-1);
		}
	}
	if((in=fopen(param.inputfile,"r"))==NULL){
		printf("\nhgt : Cannot open input file (%s)",param.inputfile);
		exit(-1);
	}
	
	//== open the bootstrapFile
	if(strcmp(param.bootstrap,"yes") == 0){
		bootstrap = 1;
	}
	if(strcmp(param.multigene,"yes") == 0){
		multigene = 1;
		if(strcmp(param.speciesroot,"file"))
			strcpy(param.speciesroot,"midpoint");
	}
	remove(param.hgtResultFile);

	initInputTree(&FirstTree);
	initInputTree(&SpeciesTreeCurrent);

	FILE * results; // = fopen(param.results,"w+");
	
	if((results = fopen(param.results,"w+"))==NULL){
		printf("PROBLEME AVEC RESULTS");
		exit(0);
	}

	
//==============================================================================
//============================= LECTURE DES ARBRES =============================
//==============================================================================
	
	printf("\nhgt : reading the input file");
	tmp = readInputFile(in, param.input/*,&SpeciesTree,&GeneTree*/,param.errorFile);

	if(tmp==-1) {
		printf("\nCannot read input data !!\n");
		exit(-1);
	}
  
	
	cpt_hgt = 0;

	initInputTree(&SpeciesTree);
	initInputTree(&GeneTree);
	initInputTree(&SpeciesTreeRed);
	initInputTree(&GeneTreeRed);

	//== lecture des matrices ou chaines newick en entree
	if(readInput(SPECIE,param.input,&SpeciesTree) == -1){ printf("\nError in species tree\n"); exit(-1);}
	if(readInput(GENE,param.input,&GeneTree) == -1){ printf("\nError in gene tree\n"); getchar(); exit(-1);}

	TrierMatrices(GeneTree.Input,GeneTree.SpeciesName,SpeciesTree.SpeciesName,SpeciesTree.size);
	
	NJ(SpeciesTree.Input,SpeciesTree.ADD,SpeciesTree.size);
	NJ(GeneTree.Input,GeneTree.ADD,GeneTree.size);

	//== construction des differentes reprsentation des arbres (adjacence,aretes,longueur,degre)
	CreateSubStructures(&SpeciesTree,1,binaireSpecies);
	CreateSubStructures(&GeneTree,1,binaireGene);

//for(i=1;i<=2*SpeciesTree.size-3-SpeciesTree.kt;i++)
	//printf("\n%d--%d : %lf",SpeciesTree.ARETE[2*i-1],SpeciesTree.ARETE[2*i-2],SpeciesTree.LONGUEUR[i-1]);

//==============================================================================
//============================ GESTION DES RACINES =============================
//==============================================================================
  //== selection de la racine
	printf("\nhgt : adding the tree roots");
	if(strcmp(param.load,"yes") == 0 ){
		//printf("On charge les fichiers");

		chargerFichier(&SpeciesTree,param.speciesTree,param.speciesRootfile);
		//addRoot(&GeneTree,NULL,GeneBranch,param.generoot,param.geneRootfile,NULL);
		chargerFichier(&GeneTree,param.geneTree,param.geneRootfile);
	}
	else{		
		if((strcmp(param.version,"web")==0) && (strcmp(param.printWeb,"yes")==0)){
		    saveTree(param.speciesTreeWeb,SpeciesTree,bestHGT,0,cpt_hgt,"",param.scenario,NULL);
		    saveTree(param.geneTreeWeb,GeneTree,bestHGT,0,cpt_hgt,"",param.scenario,NULL);
	   }

		if(first == 1){
			//FILE *froot;
			int nbBranche,leave;			
		}
		if(speciesLeaves == NULL){
			speciesLeaves = (int*) malloc(SpeciesTree.size*sizeof(int));
			speciesLeaves[0] = -1;
		}
		
		if(SpeciesTree.Root == -1) addRoot(&SpeciesTree,NULL,SpeciesBranch,param.speciesroot,param.speciesRootfile,param.speciesRootfileLeaves,NULL,param.version);
		if(GeneTree.Root == -1) addRoot(&GeneTree,NULL,GeneBranch,param.generoot,param.geneRootfile,param.geneRootfileLeaves,NULL,param.version);
	
		if(strcmp(param.viewtree,"yes")==0)
			exit(0);
	}

	nbTree++;
	if(first ==1){
		FirstTree.ADD=NULL;
		FirstTree.ARETE=NULL;
		copyInputTree(&FirstTree,SpeciesTree,1,1);
		AdjustBranchLength(&FirstTree,GeneTree,binaireSpecies,1);
	}
	
	
	SpeciesTreeCurrent.ADD=NULL;
	SpeciesTreeCurrent.ARETE=NULL;
	copyInputTree(&SpeciesTreeCurrent,SpeciesTree,1,0);

	AdjustBranchLength(&SpeciesTreeCurrent,GeneTree,binaireSpecies,1);
	AdjustBranchLength(&SpeciesTree,GeneTree,binaireSpecies,1);


	InitCriteria(&aCrit,SpeciesTree.size);
	computeCriteria(SpeciesTree.ADD,GeneTree.ADD,SpeciesTree.size,&aCrit,SpeciesTree.LONGUEUR,SpeciesTree.ARETE,GeneTree.LONGUEUR,GeneTree.ARETE);

	if(SpeciesTree.size > GeneTree.size) max_hgt = 4*GeneTree.size * GeneTree.size;
	else max_hgt = 4*SpeciesTree.size * SpeciesTree.size;

	bestHGTRed = (struct HGT*)malloc(max_hgt*sizeof(struct HGT)); 
	bestHGT = (struct HGT*)malloc(max_hgt*sizeof(struct HGT));

	for(i=0;i<max_hgt;i++){
		bestHGTRed[i].listSource = NULL;
		bestHGTRed[i].listDestination = NULL;
		bestHGT[i].listSource = NULL;
		bestHGT[i].listDestination = NULL;
	} 
	
	if(first==1 && strcmp(param.scenario,"multiple")!=0 && strcmp(param.mode,"multicheck")==0)
		multicheckTab = (struct CRITERIA *)malloc(max_hgt*sizeof(struct CRITERIA)); 

	
	if((bootstrap != 1) || (bootstrap==1 && first==1)){
		fprintf(results,"%d,%lf,%lf\n",aCrit.RF,aCrit.LS,aCrit.BD);
		RFref=aCrit.RF;
	}


	DTGene = (struct DescTree*)malloc((2*GeneTree.size-2-GeneTree.kt+1)*sizeof(struct DescTree));
	RechercherBipartition(GeneTree.ARETE,GeneTree.ADD,GeneTree.Root,GeneTree.Adjacence,DTGene,GeneTree.size,GeneTree.kt);	
	
	DTSpecies = (struct DescTree*)malloc((2*SpeciesTree.size-2-SpeciesTree.kt+1)*sizeof(struct DescTree));
	RechercherBipartition(SpeciesTree.ARETE,SpeciesTree.ADD,SpeciesTree.Root,SpeciesTree.Adjacence,DTSpecies,SpeciesTree.size,SpeciesTree.kt);

	printf("\nhgt : pre-treatment process");
	//printf("\nPre-traitement .......");
	//== date : 26 janvier 2008
	//== l'appel a cette fonction permet de recreer le meme sous arbre dans l'arbre de gene et l'arbre d'especes
	//== si les distances entre les especes sont nulles.
	int tousLesCasSontTraitees = FALSE;
	int *tab_tous_les_sommets = (int*)malloc((SpeciesTree.size+1) * sizeof(int));
	int *tab_sommets_selectionnes = (int*)malloc((SpeciesTree.size+1) * sizeof(int));
	int *tab_branches = (int*)malloc( 4*(GeneTree.size+1) * sizeof(int));
	int nb_branches;
	int temoin_nouveau_cas;
	struct HGT aHGT;
	
	for(i=1;i<=SpeciesTree.size;i++) tab_tous_les_sommets[i] = 0;
	tab_tous_les_sommets[0] = FALSE;

	while(tousLesCasSontTraitees == FALSE){
		
		ListeSommets_taille_0(GeneTree.Input,tab_tous_les_sommets,GeneTree.size-1);
		tousLesCasSontTraitees = tab_tous_les_sommets[0];
		
		if(tousLesCasSontTraitees == FALSE){
			tab_sommets_selectionnes[0] = 0;
			temoin_nouveau_cas=0;	
			for(i=1;i<GeneTree.size;i++){
				if(tab_tous_les_sommets[i] == 1){
					if(temoin_nouveau_cas == 0){
						temoin_nouveau_cas=1;
						//printf("\nTraitement : ");
					}
					tab_tous_les_sommets[i] = 2;
					tab_sommets_selectionnes[0] = tab_sommets_selectionnes[0] + 1;
					tab_sommets_selectionnes[tab_sommets_selectionnes[0]] = i;
					//printf("%d(%s) ",i,GeneTree.SpeciesName[i]);
				}
			}
			if(tab_sommets_selectionnes[0] > 0){
				ListesBranchesPourHGT(tab_sommets_selectionnes,GeneTree.ARETE,GeneTree.size,DTGene,tab_branches,&nb_branches);
				while(findBestHGT_nombreLimite(DTGene,DTSpecies,tab_branches,nb_branches,GeneTree,SpeciesTree,param,&aHGT) > 0){
					applyHGT(SpeciesTree.ADD,&GeneTree,aHGT.source,aHGT.destination);
					AdjustBranchLength(&GeneTree,SpeciesTree,0,1);
					deleteBipartition(DTGene,GeneTree);
					DTGene = (struct DescTree*)malloc((2*GeneTree.size-2-GeneTree.kt+1)*sizeof(struct DescTree));
					RechercherBipartition(GeneTree.ARETE,GeneTree.ADD,GeneTree.Root,GeneTree.Adjacence,DTGene,GeneTree.size,GeneTree.kt);	
					
					ListesBranchesPourHGT(tab_sommets_selectionnes,GeneTree.ARETE,GeneTree.size,DTGene,tab_branches,&nb_branches);
				}

			}
			
		}		
	}
	free(tab_tous_les_sommets);
	free(tab_sommets_selectionnes);
	free(tab_branches);
	
	InitCriteria(&aCrit,SpeciesTree.size);
	computeCriteria(SpeciesTree.ADD,GeneTree.ADD,SpeciesTree.size,&aCrit,SpeciesTree.LONGUEUR,SpeciesTree.ARETE,GeneTree.LONGUEUR,GeneTree.ARETE);
	//printf("\nhgt : %d %d %lf\n",aCrit.RF,SpeciesTree.size,(double)aCrit.RF/(2*(double)SpeciesTree.size-6));
	
	ReduceTree(SpeciesTree,GeneTree,&SpeciesTreeRed,&GeneTreeRed,&aMap,DTSpecies,DTGene,binaireSpecies,binaireGene);

	tmp = 0;


	//=============================== DETECTION DES TRANSFERTS ====================================
	if(strcmp(param.version,"consol")==0){
		//printf("\nDetection in progress",nbTree);		
		printf("\n===============================================");
		printf("\n| CRITERIA VALUES BEFORE DETECTION ");	
		printf("\n| RF distance  = %2d",aCrit.RF);
		printf("\n| LS criterion = %2.1lf",aCrit.LS);
		printf("\n| BD criterion = %2.1lf",aCrit.BD);
		printf("\n===============================================\n");
  }

	if(strcmp(param.scenario,"multiple")==0){
		cpt_hgt = findAllHGT(SpeciesTreeRed,GeneTreeRed,param,bestHGTRed);
		for(i=1;i<=cpt_hgt;i++){
			expandBestHGT(bestHGTRed[i],&bestHGT[i],aMap,DTSpecies,SpeciesTree);
		}
		sortHGT(bestHGT,cpt_hgt,param);
		if(cpt_hgt > param.nbhgt) cpt_hgt = param.nbhgt;
	}
	else if(strcmp(param.scenario,"unique")==0){
		if( strcmp(param.subtree,"yes") == 0 && strcmp(param.mode,"multicheck")==0 ){

			if(first==1){
				multicheckTab[0].m = 0;
				imc = 1;
			}
//				printf("\ndebut de la recherche");
      
//==============================================================================
//============ RECHERCHE DE TRANSFERTS - PLUSIEURS PAR TOUR ====================
//==============================================================================
		printf("\nhgt : start of detection");
		trivial = (SpeciesTree.kt == 0)?0:1;	
		while( findBestHGTtab(SpeciesTreeRed,GeneTreeRed,param,bestHGTRed,&nbHgtFound,&trivial,bootstrap) > 0){

			trivial = (SpeciesTree.kt == 0)?0:1;

			printf("\n\n[%d HGT%s]", nbHgtFound,(nbHgtFound > 1)?"s":"");	
			
			if(first==1){
				multicheckTab[0].m ++;
				multicheckTab[imc].nbHgtFound = nbHgtFound;
			}
			
			int temoin_zero=-1;
			int cpt_hgt2 = cpt_hgt;
			for(i=0;i<nbHgtFound;i++){
				if(bestHGTRed[i].crit.RF == 0){
					temoin_zero = i;
				}
			}
			
			for(i=0;i<nbHgtFound;i++){
				cpt_hgt++;		
				
				expandBestHGT(bestHGTRed[i],&bestHGT[cpt_hgt],aMap,DTSpecies,SpeciesTree);
				bestHGT[cpt_hgt].trivial = 0;
				bestHGTRed[i].listSource = NULL;
				bestHGTRed[i].listDestination = NULL;
				if((temoin_zero != -1)&&(i!=temoin_zero)){
					bestHGT[cpt_hgt].valide = 0; 
					continue;  
				}
				  
				if((cpt_hgt,bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_A) || 
					(cpt_hgt,bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_B) || 
					(cpt_hgt,bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_A) || 
					(cpt_hgt,bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_B)){
					bestHGT[cpt_hgt].trivial = 1;
				}
								
				if((bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_A && SpeciesTree.degre[bestHGT[cpt_hgt].source_A] == 3) ||
				   (bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_B && SpeciesTree.degre[bestHGT[cpt_hgt].source_A] == 3) ||
				   (bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_A && SpeciesTree.degre[bestHGT[cpt_hgt].source_B] == 3) ||
				   (bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_B && SpeciesTree.degre[bestHGT[cpt_hgt].source_B] == 3) ){

					bestHGT[cpt_hgt].valide = 0;
					//printf("\nuseless=> HGT #%d : [%2d--%2d] -> [%2d--%2d]",cpt_hgt,bestHGT[cpt_hgt].source_A,bestHGT[cpt_hgt].source_B,bestHGT[cpt_hgt].dest_A,bestHGT[cpt_hgt].dest_B);

					continue;
				}

				applyHGT2(GeneTree.ADD,&SpeciesTree,bestHGT[cpt_hgt].source,bestHGT[cpt_hgt].destination);
							
				if(strcmp(param.version,"consol")==0){
					printf("\nHGT #%d : [%2d--%2d] -> [%2d--%2d]",cpt_hgt,bestHGT[cpt_hgt].source_A,bestHGT[cpt_hgt].source_B,bestHGT[cpt_hgt].dest_A,bestHGT[cpt_hgt].dest_B);
				}
			}
			
			computeCriteria(SpeciesTree.ADD,GeneTree.ADD,SpeciesTree.size,&aCrit,SpeciesTree.LONGUEUR,SpeciesTree.ARETE,GeneTree.LONGUEUR,GeneTree.ARETE);
					
			printf("\n\nCriteria values after this step :");
			printf("\nRF = %d | LS = %1.2lf | BD = %1.2lf\n",aCrit.RF,aCrit.LS,aCrit.BD,aCrit.QD);	 

			if(first==1){
				multicheckTab[imc].LS = aCrit.LS;
				multicheckTab[imc].RF = aCrit.RF;
				multicheckTab[imc].BD = aCrit.BD;
				multicheckTab[imc].QD = aCrit.QD;
				imc++;
			}	
			
			if((cpt_hgt >= param.nbhgt)||(aCrit.RF == 0)) break;
					 
			deleteBipartition(DTSpecies,SpeciesTreeCurrent);
			copyInputTree(&SpeciesTreeCurrent,SpeciesTree,1,1);
			
			DTSpecies = (struct DescTree*)malloc((2*SpeciesTree.size-2-SpeciesTree.kt+1)*sizeof(struct DescTree));
			RechercherBipartition(SpeciesTree.ARETE,SpeciesTree.ADD,SpeciesTree.Root,SpeciesTree.Adjacence,DTSpecies,SpeciesTree.size,SpeciesTree.kt);
			FreeMemory_InputTreeReduced(&SpeciesTreeRed,SpeciesTreeRed.size);
			FreeMemory_InputTreeReduced(&GeneTreeRed,GeneTreeRed.size);
			initInputTree(&SpeciesTreeRed);
			initInputTree(&GeneTreeRed);
			free(aMap.map);
			free(aMap.gene);
			free(aMap.species);
			ReduceTree(SpeciesTree,GeneTree,&SpeciesTreeRed,&GeneTreeRed,&aMap,DTSpecies,DTGene,binaireSpecies,binaireGene);
					
		}
		free(aMap.map);
		free(aMap.gene);
		free(aMap.species);
		deleteBipartition(DTSpecies,SpeciesTreeCurrent);

		int retour=0;
		int cpt,nbTours,j;
		
		retour = DeleteUseLessHGT(cpt_hgt,bestHGT,SpeciesTree,FirstTree); 
			cpt=1;
			if(first==1){
				for(i=1;i<=multicheckTab[0].m;i++){
					nbTours = multicheckTab[i].nbHgtFound;
					for(j=1;j<=nbTours;j++){
						if(bestHGT[cpt].valide == 0){
							multicheckTab[i].nbHgtFound --;
						}
						cpt++;
					}
				}
			}
	}
	else{	
			int initial=1;
     			
			while(findBestHGT(initial,SpeciesTreeRed,GeneTreeRed,param,&bestHGTRed[cpt_hgt+1]) > 0){
				cpt_hgt++;		
				expandBestHGT(bestHGTRed[cpt_hgt],&bestHGT[cpt_hgt],aMap,DTSpecies,SpeciesTree);
              	bestHGT[cpt_hgt].trivial = 0;
          		if((cpt_hgt,bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_A) || 
             		(cpt_hgt,bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_B) || 
             		(cpt_hgt,bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_A) || 
             		(cpt_hgt,bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_B)){
              		bestHGT[cpt_hgt].trivial = 1;
          		}	
				if((bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_A && SpeciesTree.degre[bestHGT[cpt_hgt].source_A] == 3) ||
					   (bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_B && SpeciesTree.degre[bestHGT[cpt_hgt].source_A] == 3) ||
					   (bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_A && SpeciesTree.degre[bestHGT[cpt_hgt].source_B] == 3) ||
					   (bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_B && SpeciesTree.degre[bestHGT[cpt_hgt].source_B] == 3) ){

						bestHGT[cpt_hgt].valide = 0;
						if(strcmp(param.version,"consol")==0){
							//printf("\nuseless=> HGT #%d : [%2d--%2d] -> [%2d--%2d]",cpt_hgt,bestHGT[cpt_hgt].source_A,bestHGT[cpt_hgt].source_B,bestHGT[cpt_hgt].dest_A,bestHGT[cpt_hgt].dest_B);
						}
					continue;
				}

				bestHGTRed[i].listSource = NULL;
				bestHGTRed[i].listDestination = NULL;
				
				applyHGT(GeneTree.ADD,&SpeciesTree,bestHGT[cpt_hgt].source,bestHGT[cpt_hgt].destination);
				AdjustBranchLength(&SpeciesTree,GeneTree,0,1);
				
				computeCriteria(SpeciesTree.ADD,GeneTree.ADD,SpeciesTree.size,&aCrit,SpeciesTree.LONGUEUR,SpeciesTree.ARETE,GeneTree.LONGUEUR,GeneTree.ARETE);
				loadCriteria(aCrit,&(bestHGT[cpt_hgt]));
				
				if(strcmp(param.version,"consol")==0){
					printf("\nHGT #%d %d--%d -> %d--%d",cpt_hgt,bestHGT[cpt_hgt].source_A,bestHGT[cpt_hgt].source_B,bestHGT[cpt_hgt].dest_A,bestHGT[cpt_hgt].dest_B);
					printf("\nRF = %d, LS = %lf, BD = %lf \n",aCrit.RF,aCrit.LS,aCrit.BD,aCrit.QD);	 
				} 
				if(bestHGT[cpt_hgt].crit.RF == 0) break;// || bestHGT[cpt_hgt].crit.LS < epsilon) break;

				if(cpt_hgt >= param.nbhgt) break;
			
				deleteBipartition(DTSpecies,SpeciesTreeCurrent);
				copyInputTree(&SpeciesTreeCurrent,SpeciesTree,1,1);
				DTSpecies = (struct DescTree*)malloc((2*SpeciesTree.size-2-SpeciesTree.kt+1)*sizeof(struct DescTree));
				RechercherBipartition(SpeciesTree.ARETE,SpeciesTree.ADD,SpeciesTree.Root,SpeciesTree.Adjacence,DTSpecies,SpeciesTree.size,SpeciesTree.kt);
				
				free(aMap.map);
				free(aMap.gene);
				free(aMap.species);
				FreeMemory_InputTreeReduced(&SpeciesTreeRed,SpeciesTreeRed.size);
				FreeMemory_InputTreeReduced(&GeneTreeRed,GeneTreeRed.size);
				initInputTree(&SpeciesTreeRed);
				initInputTree(&GeneTreeRed);
					
				ReduceTree(SpeciesTree,GeneTree,&SpeciesTreeRed,&GeneTreeRed,&aMap,DTSpecies,DTGene,binaireSpecies,binaireGene);
				//printf("<br> prochlorococcus = %d", aMap.species[15]);
			}
			
			initial=0; 
			while( findBestHGT(initial,SpeciesTreeRed,GeneTreeRed,param,&bestHGTRed[cpt_hgt+1]) > 0){	
				cpt_hgt++;		
				expandBestHGT(bestHGTRed[cpt_hgt],&bestHGT[cpt_hgt],aMap,DTSpecies,SpeciesTree);
              	bestHGT[cpt_hgt].trivial = 0;
          		if((cpt_hgt,bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_A) || 
             		(cpt_hgt,bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_B) || 
             		(cpt_hgt,bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_A) || 
             		(cpt_hgt,bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_B)){
              		bestHGT[cpt_hgt].trivial = 1;
          		}	
					if((bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_A && SpeciesTree.degre[bestHGT[cpt_hgt].source_A] == 3) ||
					   (bestHGT[cpt_hgt].source_A == bestHGT[cpt_hgt].dest_B && SpeciesTree.degre[bestHGT[cpt_hgt].source_A] == 3) ||
					   (bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_A && SpeciesTree.degre[bestHGT[cpt_hgt].source_B] == 3) ||
					   (bestHGT[cpt_hgt].source_B == bestHGT[cpt_hgt].dest_B && SpeciesTree.degre[bestHGT[cpt_hgt].source_B] == 3) ){

						bestHGT[cpt_hgt].valide = 0;
						if(strcmp(param.version,"consol")==0){
							printf("\nuseless=> HGT #%d : [%2d--%2d] -> [%2d--%2d]",cpt_hgt,bestHGT[cpt_hgt].source_A,bestHGT[cpt_hgt].source_B,bestHGT[cpt_hgt].dest_A,bestHGT[cpt_hgt].dest_B);
						}
						//printf("\nTransfert %d inutile ",cpt_hgt);
						continue;
					}

				bestHGTRed[i].listSource = NULL;
				bestHGTRed[i].listDestination = NULL;
					
			//printf("\navant apply dans findBestHGT");
				applyHGT(GeneTree.ADD,&SpeciesTree,bestHGT[cpt_hgt].source,bestHGT[cpt_hgt].destination);
			//printf("\napres apply dans findBestHGT");
			
			//printf("\ntoto");
				AdjustBranchLength(&SpeciesTree,GeneTree,0,1);
			
			/*for(i=1;i<=2*GeneTree.size-3;i++)
				printf("\n%d-%d -->%lf",GeneTree.ARETE[2*i-1],GeneTree.ARETE[2*i-2],GeneTree.LONGUEUR[i-1]);*/
				
			
			computeCriteria(SpeciesTree.ADD,GeneTree.ADD,SpeciesTree.size,&aCrit,SpeciesTree.LONGUEUR,SpeciesTree.ARETE,GeneTree.LONGUEUR,GeneTree.ARETE);
			loadCriteria(aCrit,&(bestHGT[cpt_hgt]));
			//printf("\navant Delete");
				//if (cpt_hgt > 1)
				//	DeleteUseLessHGT(cpt_hgt,bestHGT,SpeciesTree,FirstTree);
				/*	
					int retour;
					do{						
						printf("\ntest des transferts inutiles");
						retour = DeleteUseLessHGT(cpt_hgt,bestHGT,SpeciesTree,FirstTree); 
						printf("\nretour = %d",retour);
					}while(retour > 0);
				}*/
			//printf("\napres Delete");
			//printf("ici = %s, %d",param.version,cpt_hgt);
				if(strcmp(param.version,"consol")==0){
					printf("\nHGT #%d %d--%d -> %d--%d",cpt_hgt,bestHGT[cpt_hgt].source_A,bestHGT[cpt_hgt].source_B,bestHGT[cpt_hgt].dest_A,bestHGT[cpt_hgt].dest_B);
					printf("\nRF = %d, LS = %lf, BD = %lf\n",aCrit.RF,aCrit.LS,aCrit.BD,aCrit.QD);	 
				}
			//printf("icitttt = %d , %d",bestHGT[cpt_hgt].crit.RF,param.nbhgt);
				if(bestHGT[cpt_hgt].crit.RF == 0) {
					int retour;
					do{						
					//	printf("\ntest des transferts inutiles");
						retour = DeleteUseLessHGT(cpt_hgt,bestHGT,SpeciesTree,FirstTree); 
					//	printf("\nretour = %d",retour);
					}while(retour > 0);
				
					break;// || bestHGT[cpt_hgt].crit.LS < epsilon) break;
				}
				if(cpt_hgt >= param.nbhgt) break;
			
				deleteBipartition(DTSpecies,SpeciesTreeCurrent);
				copyInputTree(&SpeciesTreeCurrent,SpeciesTree,1,1);
				free(aMap.map);
				free(aMap.gene);
				free(aMap.species);
				DTSpecies = (struct DescTree*)malloc((2*SpeciesTree.size-2-SpeciesTree.kt+1)*sizeof(struct DescTree));
				RechercherBipartition(SpeciesTree.ARETE,SpeciesTree.ADD,SpeciesTree.Root,SpeciesTree.Adjacence,DTSpecies,SpeciesTree.size,SpeciesTree.kt);
				FreeMemory_InputTreeReduced(&SpeciesTreeRed,SpeciesTreeRed.size);
				FreeMemory_InputTreeReduced(&GeneTreeRed,GeneTreeRed.size);
				initInputTree(&SpeciesTreeRed);
				initInputTree(&GeneTreeRed);
			
				ReduceTree(SpeciesTree,GeneTree,&SpeciesTreeRed,&GeneTreeRed,&aMap,DTSpecies,DTGene,binaireSpecies,binaireGene);
			//printf("==>%d",SpeciesTreeRed.size);
				//printf("<br> prochlorococcus = %d", aMap.map[6]);
		}
		free(aMap.map);
		free(aMap.gene);
		free(aMap.species);
		deleteBipartition(DTSpecies,SpeciesTreeCurrent);
		//FreeMemory_InputTreeReduced(&SpeciesTreeRed,SpeciesTreeRed.size);
		//FreeMemory_InputTreeReduced(&GeneTreeRed,GeneTreeRed.size);
	   }
	}
	
//==============================================================================
//============================ TRAITEMENT DES RESULTATS ========================
//==============================================================================
	printf("\nhgt : formatting the results");
	outHGT = (struct HGT*)malloc(2*param.nbhgt*sizeof(struct HGT)); 
	nbHGT = formatResult(bestHGT,cpt_hgt,outHGT,FirstTree);
	printHGT(results,multicheckTab,param.mode,RFref,/*out,*/FirstTree,outHGT,nbHGT,NULL,param.subtree,param.bootmin);

	if((strcmp(param.version,"web")==0) && (strcmp(param.printWeb,"yes")==0)){
		saveTree(param.outputWeb,FirstTree,outHGT,1,nbHGT,param.subtree,param.scenario,NULL);
	}
		
//==============================================================================
//============================ LIBERATION DE LA MEMOIRE ========================
//==============================================================================		
	deleteBipartition(DTGene,GeneTree);
	FreeMemory_InputTreeReduced(&SpeciesTreeRed,SpeciesTreeRed.size);
	FreeMemory_InputTreeReduced(&GeneTreeRed,GeneTreeRed.size);
	FreeMemory_InputTree(&SpeciesTreeCurrent,SpeciesTreeCurrent.size);
	FreeMemory_InputTree(&GeneTree,GeneTree.size);
	FreeMemory_InputTree(&SpeciesTree,SpeciesTree.size);
	if(bootstrap !=1)
		FreeMemory_InputTree(&FirstTree,FirstTree.size);
	FreeCriteria(&aCrit,SpeciesTree.size);
	
	for(i=1;i<=cpt_hgt;i++){
		free(bestHGT[i].listSource);
		free(bestHGT[i].listDestination);
	}
	free(bestHGTRed);
	free(bestHGT); 
	
  fclose(results);
 
 //printf("\nvaleur retournee : %d \n",nbHGT);
  printf("\nhgt : number of HGT(s) found = %d \nhgt : end of computation, check the file results.txt for the program output\n",nbHGT);
  exit(nbHGT);
}
