/*
 *  Armadillo Workflow Platform v1.0
 *  A simple pipeline system for phylogenetic analysis
 *  
 *  Copyright (C) 2009-2011  Etienne Lord, Mickael Leclercq
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package biologic;


import biologic.seqclasses.parserNewick.newick_tree;
import biologic.seqclasses.parserNewick.node;
import configuration.Config;
import configuration.Util;
import database.databaseFunction;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import workflows.workflow_properties;
 

public class MultipleTrees implements Biologic, Serializable {
     private Vector<Tree> trees=new Vector<Tree>();
     private int id=0;
     private int alignment_id=0;
     private int multiplesequences_id=0;
     private String name="UnknownMultipleTrees";
     private String note="";
     private int runProgram_id=0;
     public static databaseFunction df=new databaseFunction();

public MultipleTrees() {}

public MultipleTrees(int id) {
    this.loadFromDatabase(id);
}

public MultipleTrees(String filename) {
    loadFromFile(filename);
}

public boolean loadFromFile(String filename) {
    if (!Util.FileExists(filename)) return false;
    if (!readNewick(filename)) {             
             return readNexus(filename);
         }      
     return true;   
}

////////////////////////////////////////////////////////////////////////////////
/// Database function

    public boolean loadFromDatabase(int id) {
        MultipleTrees tree=df.getMultipleTrees(id);
        if (tree.getId()>0) {
            this.setId(tree.getId());
            this.setName(tree.getName());
            this.setNote(tree.getNote());
            this.setRunProgram_id(tree.getRunProgram_id());
             this.trees.clear();
             this.trees.addAll(tree.getTree());
            return true;
        } else return false;
    }

public boolean saveToDatabase() {
    id=0;
    return (df.addMultipleTrees(this)==0?false:true);
 }

////////////////////////////////////////////////////////////////////////////////
/// Iterator

    Vector<Integer>next=new Vector<Integer>();
    int counter=0;
    int maxid=-1;
    public boolean hasNext() {
       //next=df.getAllSequenceID();
       if (next.size()==0) {
           next=df.getAllMultipleTreesID();
           maxid=next.size();
       }
       return (this.counter<maxid);
    }

    public Object next() {
        return new MultipleTrees(next.get(counter++));
    }

      @Override
    public Vector<Integer> getAllId() {
        return next;
    }

    public boolean exists(Integer id) {
        return df.existsMultipleTrees(id);
    }

    ////////////////////////////////////////////////////////////////////////////
    /// Database Function

    public void remove() {
        MultipleTrees s=new MultipleTrees(counter-1);
        s.removeFromDatabase();
    }
 public boolean readNewick(String filename) {
      if (Config.library_mode) System.out.println("Read Nexus tree from "+filename); 
     try {
          filename=filename.trim();
          BufferedReader br =new BufferedReader(new FileReader(new File(filename)));
          String Treea="";
          int index_tree=0;
          while (br.ready()) {
              String stri=br.readLine();
              Treea+=stri.trim();
              if (stri.startsWith("#NEXUS")) {                  
                   if (Config.library_mode) System.out.println("Wrong. Tree is Nexus");
                  //--Wrong read nexus..                  
                  return false;
              }
              
              if (stri.endsWith(";")) {
                  int index=Treea.indexOf("[");
                  if (index>-1) Treea=Treea.substring(0,index); //Remove annotation
                  Treea=Treea.replaceAll(";", "");              //Remove ; Ending
                  Tree t=new Tree();
                  t.setName(name+"_"+index_tree++);
                  t.setNote("From "+filename+" on "+Util.returnCurrentDateAndTime());
                  t.setTree(Treea+";");
                  t.setTreeAbbreviate(Treea+";");
                  t.setTreeSequenceID(Treea+";");
                  t.replaceSequenceIdWithNames();
                  this.trees.add(t);
                  Treea=""; //--Remove buffer
              }
          }
          br.close();
          this.replaceSequenceIDwithNames();
          return true;
      } catch(Exception e) {          
          return false;}
  }
 
  /**
     * Warning, read nexus tree but only the first one
     * *Better use MultipleTrees.readNexus(String filename)
     * See http://hydrodictyon.eeb.uconn.edu/eebedia/index.php/Phylogenetics:_NEXUS_Format
     * and http://molecularevolution.org/resources/treeformats for format
     * @param filename
     * @return
     */
    public boolean readNexus(String filename) {
      if (Config.library_mode) System.out.println("Read Nexus tree(s) from "+filename);
      try {
         BufferedReader br =new BufferedReader(new FileReader(new File(filename)));
          String Treea="";
          boolean nexus_flag=false;
          boolean inside_tree_block=false;
          Pattern p_taxon=Pattern.compile("([0-9]{1,})\\s+(\\S+)[,;]", Pattern.CASE_INSENSITIVE);
          Pattern p_tree=Pattern.compile("tree\\s+(\\S+)\\s+[=]+\\s+(.*)", Pattern.CASE_INSENSITIVE);
          Pattern p_parameters=Pattern.compile("\\[(.*?)\\]");
          ArrayList<String>parameters=new ArrayList<String>();
          HashMap<String,String>taxon_names=new HashMap<String,String>();
                    
          while (br.ready()) {
              String stri=br.readLine();             
              if (stri.startsWith("#NEXUS")) {
                  nexus_flag=true;
              }
              if (stri.startsWith("begin trees;")) {
                  //-- debug System.out.println("Inside trees");
                  inside_tree_block=true;
              }
              if (stri.startsWith("end;")) {
                  taxon_names.clear();
                  
                  inside_tree_block=false;
              }
              if (nexus_flag&&inside_tree_block) {
                  if (stri.startsWith("translate")) {} //--Do nothing
                  Matcher m_taxon=p_taxon.matcher(stri);
                  Matcher m_tree=p_tree.matcher(stri);
                  if (m_tree.find()) {
                      //-- debug System.out.println("tree find "+m_tree.group());
                      parameters.clear();
                      Treea=m_tree.group(2);
                      
                      //--Remove the parameters from tree...
                      Matcher m_parameters=p_parameters.matcher(Treea);
                      while (m_parameters.find()) {
                          parameters.add(m_parameters.group(1));
                      }                      
                      for (String param:parameters) Treea=Treea.replaceAll("\\["+param+"\\]", "");                      
                      //--Replace all taxon (number)->name...
                      newick_tree n=new newick_tree();
                      n.parseNewick(Treea);
                      for(node nod:n.node_list) {
                         if (nod.isleaf&&(taxon_names.containsKey(nod.getName()))) {
                             nod.setName(taxon_names.get(nod.getName()));
                         } 
                      }
                      Tree tmp=new Tree();
                      tmp.setNote("Loaded from "+filename);
                      tmp.setTree(n.PrintNewick());
                      tmp.replaceSequenceIdWithNames();
                      tmp.setName(m_tree.group(1));                       
                      this.trees.add(tmp);
                  } else  
                  if (m_taxon.find()) {                     
                      //-- debug System.out.println(m_taxon.group());
                      taxon_names.put(m_taxon.group(1), m_taxon.group(2));
                  }   
                  
              } //--End nexus block
             
          }
           br.close();
      } catch(Exception e) {e.printStackTrace();return false;}
      return true;
  }
 
 /**
  * Note: special function to read NewickTree from MrBayes
  * @author Med Amine Remita 
  * @since Mars 2011
  * @param filename
  * @return 
  */
    public boolean readNewickTreeFromMrBayes(String filename){
        try {
         BufferedReader br =new BufferedReader(new FileReader(new File(filename)));
         String Tree1="";
         String Tree2="";
          while (br.ready()) {
              String stri=br.readLine();
              //1er arbre
              Tree1=stri.trim();
              if (Tree1.contains("tree con_50_majrule")) {
                  int index1=Tree1.indexOf("(");
                  if (index1>-1) Tree1=Tree1.substring(index1);
                  this.trees.add(createTreeFromMrBayes(Tree1,filename));       //fonction is below
                  //2eme arbre
                  while (br.ready()){
                      stri=br.readLine();
                      Tree2=stri.trim();
                      if(Tree2.contains("tree con_50_majrule")){
                            int index2=Tree2.indexOf("(");
                            if (index2>-1) Tree2=Tree2.substring(index2);
                            this.trees.add(createTreeFromMrBayes(Tree2,filename));
                      }
                  }
              }
          }
          br.close();
          return true;
      } catch(Exception e) {return false;}
  }
    
  /**
  * Helper function to readNewickTreeFromMrBayes
  * @author Med Amine Remita 
  * @since Mars 2011
  * @param filename
  * @return 
  */
    private Tree createTreeFromMrBayes (String NewickTree, String filename){
        Tree treea = new Tree ();
                  treea.setNote("From "+filename+" on "+Util.returnCurrentDateAndTime());
                  treea.setTree(NewickTree);
                  treea.setTreeAbbreviate(NewickTree);
                  treea.setTreeSequenceID(NewickTree);
                  treea.replaceSequenceIdWithNames();
                  return treea;
    }


    /**
     * @return the tree
     */
    public Vector<Tree> getTree() {
        return trees;
    }


    public int getNbTrees() {
        return trees.size();
    }

    public boolean add(Tree tree) {
        return trees.add(tree);
    }

    /**
     * @return the alignment_id
     */
    public int getAlignment_id() {
        return alignment_id;
    }

    /**
     * @param alignment_id the alignment_id to set
     */
    public void setAlignment_id(int alignment_id) {
        this.alignment_id = alignment_id;
        for (Tree tree:trees) {
            tree.setAlignment_id(this.alignment_id);
        }
    }

    /**
     * @return the runProgram_id
     */
    public int getRunProgram_id() {
        return runProgram_id;
    }

    public void setNote(String note) {
        this.note=note;
    }

    /**
     * @param runProgram_id the runProgram_id to set
     */
    public void setRunProgram_id(int runProgram_id) {
        this.runProgram_id = runProgram_id;
        for (Tree tree:trees) tree.setRunProgram_id(runProgram_id);
    }

    public void replaceSequenceIDwithNames() {
        for (Tree tree:trees) tree.replaceSequenceIdWithNames();
    }

    public void checkNewick() {
        for (Tree tree:trees) tree.removeBootstrap();
    }
     
    @Override
    public String toString() {
        String s="MultipleTrees with "+trees.size()+" tree(s)\n";
        for (Tree t:trees) s+=t+"\n";
        return s;
    }

     public boolean outputTrees (String filename) {
        try {
            PrintWriter pw = new PrintWriter(new FileWriter(new File(filename)));
            //Ugly, need better id of trees
            for (Tree st:this.trees) pw.println(st.getTreeSequenceID()+";");
            pw.flush();
            pw.close();
        } catch (Exception e) {return false;}
     return true;
    }

    /**
     * @return the name
     */
    public String getName() {
        return name;
    }

    /**
     * @param name the name to set
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * @return the note
     */
    public String getNote() {
        return note;
    }

    /**
     * @return the id
     */
    public int getId() {
        return id;
    }

    /**
     * @param id the id to set
     */
    public void setId(int id) {
        this.id = id;
    }

    /**
     * @return the multiplesequences_id
     */
    public int getMultiplesequences_id() {
        return multiplesequences_id;
    }

    /**
     * @param multiplesequences_id the multiplesequences_id to set
     */
    public void setMultiplesequences_id(int multiplesequences_id) {
        this.multiplesequences_id = multiplesequences_id;
        for (Tree tree:trees) {
            tree.setMultiplesequences_id(multiplesequences_id);
        }
    }

    public boolean removeFromDatabase() {
        return df.removeMultipleTrees(this);
    }

    public boolean updateDatabase() {
        return this.updateDatabase();
    }

    public workflow_properties returnProperties() {
        throw new UnsupportedOperationException("Not supported yet.");
    }

    public String getBiologicType() {
        return "MultipleTrees";
    }

    public String toHtml() {
        return toString();
    }

    public String getNameId(int id) {
        return df.getMultipleTreesName(id);
    }

     public void setData(String data) {
        this.getTree().clear();
        Util u=new Util("temp.txt");
        u.println(data);
        u.close();
        this.readNewick("temp.txt");
        this.replaceSequenceIDwithNames();
        Util.deleteFile("temp.txt");
    }

     public String getFasta() {
        return "";
    }

    public String getPhylip() {
        return "";
    }

    public String getExtendedString() {
        return toString();
    }
    
     /**
     * Get the next valid Id found in the database
     * @param start the current id...
     * @return the next id or 0 if not found
     */
    public int getNextValidId() {
        if (!hasNext()) return 0;
        for (int i=0;i<next.size();i++) {
            if (next.get(i)==this.id) return next.get(i+1);
        }
        //--Unable to find the current ID? -> Find the first higher...
         for (int i=0;i<next.size();i++) {
            if (next.get(i)>this.id) return next.get(i);
        }
        //--If not, return 0; 
        return 0;
    }
}
